# -*- coding: utf-8 -*-
"""

Author
------
Bo Zhang

Email
-----
bozhang@nao.cas.cn

Created on
----------
- Sat Sep 03 16:00:00 2017

Modifications
-------------
- Sat Sep 03 12:00:00 2017

Aims
----
- plotting tools

"""


import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import binned_statistic_2d
from matplotlib import cm
from lmfit.models import GaussianModel
from mpl_toolkits.mplot3d import Axes3D

from .analysis import label_diff_lmfit


def plot_mse(s):
    fig, ax = plt.subplots(1, 1, figsize=(8, 6), tight_layout=True)
    plt.hist(-s.nmse[s.nmse != 0], np.linspace(0, 1, 80), histtype='step',
             lw=2, label="MSE")
    plt.hist(-s.scores[s.nmse != 0], np.linspace(0, 1, 80), histtype='step',
             lw=2, label="CV MSE")
    ylim = plt.gca().get_ylim()
    plt.vlines(np.percentile(-s.nmse[s.nmse != 0], [14, 50, 86]), *ylim,
               linestyle='--', label="14, 50, 86 percentiles")
    plt.xlim(0, 1)
    plt.ylim(*ylim)
    plt.ylabel("Counts")
    plt.xlabel("MSE")
    fig.tight_layout()
    return fig


# ################ #
# image
# ################ #
def image(ax, x, y, xbins, ybins, log=True):
    plt.sca(ax)
    c, xe, ye, bn = binned_statistic_2d(x, y, x, statistic="count",
                                        bins=[xbins, ybins])
    if log:
        c = np.log10(c)
    plt.imshow(c.T, origin="lower", extent=(*xbins[[0, -1]], *ybins[[0, -1]]),
               cmap=cm.viridis, aspect="auto")
    return


def compare_labels(X_true, X_pred,
                   xlabels=None, ylabels=None, reslabels=None,
                   xlims=None, reslims=None,
                   histlim=None, nxb=30, cornerlabel="",
                   figsize=None):

    nlabel = X_true.shape[1]

    if xlabels is None:
        xlabels = ["$X_{{true}}:{}$".format(i) for i in range(nlabel)]
    if ylabels is None:
        ylabels = ["$X_{{pred}}:{}$".format(i) for i in range(nlabel)]
    if reslabels is None:
        reslabels = ["$X_{{res}}:{}$".format(i) for i in range(nlabel)]

    # default xlim
    if xlims is None:
        xlim1 = np.min(np.vstack((np.percentile(X_true, 1, axis=0),
                                  np.percentile(X_pred, 1, axis=0))), axis=0)
        xlim2 = np.min(np.vstack((np.percentile(X_true, 99, axis=0),
                                  np.percentile(X_pred, 99, axis=0))), axis=0)
        xlims = (xlim2 - xlim1).reshape(-1, 1) * 0.4 * np.array(
            [-1, 1]) + np.vstack((xlim1, xlim2)).T
    if reslims is None:
        reslims = np.repeat(
            np.max(np.abs(np.percentile(X_pred - X_true, [1, 99], axis=0).T),
                   axis=1).reshape(-1, 1), 2, axis=1) * np.array([-1, 1])
        reslims = np.abs(np.diff(reslims, axis=1)) * np.array(
            [-1, 1]) * 0.2 + reslims

    # run MCMC
    X_bias, X_scatter, frs, histdata = label_diff_lmfit(
        X_true, X_pred, bins="auto", plot=False, emcee=True)
    print("bias", X_bias)
    print("scatter", X_scatter)
    if histlim is None:
        histlim = (0, np.max([np.max(histdata_[0]) for histdata_ in histdata]))
    histlim = np.array(histlim)

    if figsize is None:
        figsize = (3 * nlabel, 3 * nlabel)

    # draw figure
    fig, axs2 = plt.subplots(nlabel+1, nlabel+1, figsize=figsize)

    # 1. Gaussian
    gm = GaussianModel()
    for i in range(nlabel):
        plt.sca(axs2[i + 1, -1])
        fr = frs[i]
        hist_, bin_edge_, data_ = histdata[i]
        plt.hist(data_, bins=bin_edge_, histtype="step",
                 orientation="horizontal")
        axs2[i + 1, -1].plot(gm.eval(fr.mcmc.params, x=bin_edge_), bin_edge_)
        axs2[i + 1, -1].tick_params(direction='in', pad=5)
        axs2[i + 1, -1].set_xlim(histlim)
        axs2[i + 1, -1].set_ylim(reslims[i])
        axs2[i + 1, -1].set_ylim(reslims[i])
        axs2[i + 1, -1].yaxis.tick_right()
        axs2[i + 1, -1].hlines(X_bias[i], *histlim, linestyle='--', color="k")

        pos_text_x = np.dot(np.array([[0.9, 0.1]]), histlim.reshape(-1, 1))
        pos_text_y = np.dot(np.array([[0.15, 0.85]]),
                            reslims[i].reshape(-1, 1))
        axs2[i + 1, -1].text(pos_text_x, pos_text_y,
                             "$bias={:.4f}$".format(X_bias[i]))
        pos_text_x = np.dot(np.array([[0.9, 0.1]]), histlim.reshape(-1, 1))
        pos_text_y = np.dot(np.array([[0.30, 0.70]]),
                            reslims[i].reshape(-1, 1))
        axs2[i + 1, -1].text(pos_text_x, pos_text_y,
                             "$\\sigma={:.4f}$".format(X_scatter[i]))

        axs2[i + 1, -1].yaxis.tick_right()

        if i < nlabel-1:
            axs2[i + 1, -1].set_xticklabels([])

    axs2[-1, -1].set_xlabel("Counts")

    # 2. diagnal
    for i in range(nlabel):
        image(axs2[0, i], X_true[:, i], X_pred[:, i],
              np.linspace(xlims[i][0], xlims[i][1], nxb),
              np.linspace(xlims[i][0], xlims[i][1], nxb))
        axs2[0, i].set_xlim(*xlims[i])
        axs2[0, i].set_ylim(*xlims[i])
        axs2[0, i].tick_params(direction='in', pad=5)
        axs2[0, i].set_xticklabels([])
        axs2[0, i].set_ylabel(ylabels[i])
        axs2[0, i].plot(xlims[i], xlims[i], 'k--')

    # 3. Xres vs X
    X_res = X_pred - X_true
    for i in range(nlabel):
        for j in range(nlabel):
            image(axs2[j + 1, i], X_true[:, i], X_res[:, j],
                  np.linspace(xlims[i][0], xlims[i][1], nxb),
                  np.linspace(reslims[j][0], reslims[j][1], nxb))
            axs2[j + 1, i].set_xlim(*xlims[i])
            axs2[j + 1, i].set_ylim(*reslims[j])
            axs2[j + 1, i].tick_params(direction='in', pad=5)

            if j != nlabel - 1:
                axs2[j + 1, i].set_xticklabels([])
            else:
                axs2[j + 1, i].set_xlabel(xlabels[i])

            if i != 0:
                axs2[j + 1, i].set_yticklabels([])
            else:
                axs2[j + 1, i].set_ylabel(reslabels[j])

    axs2[0, -1].set_axis_off()
    axs2[0, -1].text(np.mean(axs2[0, -1].get_xlim()),
                     np.mean(axs2[0, -1].get_ylim()),
                     cornerlabel, horizontalalignment='center',
                     verticalalignment='center')

    fig.tight_layout()
    plt.subplots_adjust(wspace=0., hspace=0.)

    return fig, frs

# gs1 = plt.GridSpec(1, 6, left=0.05, bottom=0.85, right=0.95, top=0.95, hspace=0.2, wspace=0)
# gs2 = plt.GridSpec(7, 7, left=0.05, bottom=0.1, right=0.95, top=0.8, hspace=0, wspace=0)

# axs1 = np.array([fig.add_subplot(gs1[i]) for i in range(6)])
# axs2 = np.array([[fig.add_subplot(gs2[j, i]) for i in range(7)] for j in range(7)])


def visual3d(s, wave, diag_dim=()):
    """ single pixel diagnostic """

    diag_dim = (0, 1)

    i_pixel = 5175 - 3900
    # i_pixel = 4861-3900
    x, y, flux = s.single_pixel_diagnostic(i_pixel, s.tr_labels,
                                           diag_dim=diag_dim)

    fig = plt.figure(figsize=(10, 10))
    ax = fig.add_subplot(111, projection='3d')
    ax.scatter3D(s.tr_labels[:, 0], s.tr_labels[:, 1], s.tr_flux[:, i_pixel],
                 s=10, c=s.tr_labels[:, 2], alpha=.5, cmap=cm.jet)
    plt.colorbar()
    # plt.plot(x,y, flux,'b.')
    ax.set_zlim(0., 2.)
    plt.xlabel('Teff')
    plt.ylabel('logg')
    plt.title('PIXEL: %s' % i_pixel)
    fig.tight_layout()

    # %%
    diag_dim = (0, 2)

    i_pixel = 6564 - 3900
    i_pixel = 4861 - 3900

    sdiag_teff = np.arange(4000., 8000., 100.)
    sdiag_logg = np.arange(1., 5., .2)
    sdiag_logg = np.arange(-2, 1., .2)
    msdiag_teff, msdiag_logg = np.meshgrid(sdiag_teff, sdiag_logg)
    msdiag_feh = np.zeros_like(msdiag_teff)
    msdiag_labels = np.array([msdiag_teff.flatten(),
                              msdiag_logg.flatten(),
                              msdiag_feh.flatten()]).T

    H, xedges, yedges = np.histogram2d(train_labels[:, 0], train_labels[:, 2],
                                       bins=(sdiag_teff, sdiag_logg))
    xcenters = (xedges[:-1] + xedges[1:]) / 2.
    ycenters = (yedges[:-1] + yedges[1:]) / 2.
    xmesh, ymesh = np.meshgrid(xcenters, ycenters)

    x, y, flux = s.single_pixel_diagnostic(i_pixel, msdiag_labels,
                                           diag_dim=diag_dim)

    # flux[:20]=np.nan

    fig = plt.figure(figsize=(10, 10))
    ax = fig.add_subplot(111, projection='3d')
    # ax.scatter(x,y,flux, c=msdiag_feh)
    surf = ax.plot_surface(msdiag_teff, msdiag_logg,
                           flux.reshape(msdiag_teff.shape),
                           vmin=np.min(flux), vmax=np.max(flux), cmap=cm.jet)
    ax.contour(xmesh, ymesh, np.log(H.T), extend3d=False, offset=1.20,
               color='k')
    # plt.plot(x,y, flux,'b.')
    ax.set_zlim(0., 2.)
    plt.xlabel('Teff')
    plt.ylabel('logg')
    plt.title('PIXEL: %s' % i_pixel)

    fig.colorbar(surf, shrink=.5, aspect=5)
    fig.tight_layout()
    # fig.savefig("../data/laap/figs/PIXEL6564_C8_E0P08_G1.svg")


import numpy as np

"""

define error processing functions

"""


# jacobian matrix to covariance
def jac_to_cov(jac):
    from scipy.linalg import svd
    _, s, VT = svd(jac, full_matrices=False)
    threshold = np.finfo(float).eps * max(jac.shape) * s[0]
    s = s[s > threshold]
    VT = VT[:s.size]
    pcov = np.dot(VT.T / s**2, VT)
    return pcov


# covariance to standard deviation
def cov_to_err(cov):
    return np.sqrt(np.diag(cov))


# jacobian matrix to standard deviation
def jac_to_err(jac):
    from scipy.linalg import svd
    _, s, VT = svd(jac, full_matrices=False)
    threshold = np.finfo(float).eps * max(jac.shape) * s[0]
    s = s[s > threshold]
    VT = VT[:s.size]
    pcov = np.dot(VT.T / s**2, VT)
    return np.sqrt(np.diag(pcov))


# inversed hessian matrix to standard deviation
def hessinv_to_err(hess_inv):
    return np.sqrt(np.diag(hess_inv))


# least_squares[jac] --> jacobian matrix --> standard deviation
def do_post(ls_result, label_scaler=None):
    """

    Parameters
    ----------
    ls_result:
        least_squares full output
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.least_squares.html

    label_scaler:
        SLAM label scale for training labels

    Returns
    -------
    pp_result:
        post-processed result

    """
    pp_result = dict()
    pp_result["cost"] = ls_result["cost"]
    pp_result["grad"] = ls_result["grad"]

    pp_result["pcov"] = jac_to_cov(ls_result["jac"]) * \
                        np.dot(label_scaler.scale_.reshape(-1, 1), label_scaler.scale_.reshape(1, -1))
    pp_result["pstd"] = cov_to_err(pp_result["pcov"])

    pp_result["message"] = ls_result["message"]
    pp_result["nfev"] = ls_result["nfev"]
    pp_result["optimality"] = ls_result["optimality"]
    pp_result["status"] = ls_result["status"]
    pp_result["success"] = ls_result["success"]

    pp_result["x"] = label_scaler.inverse_transform(ls_result["x"].reshape(1, -1)).flatten()

    return pp_result



# -*- coding: utf-8 -*-
"""

Author
------
Bo Zhang

Email
-----
bozhang@nao.cas.cn

Created on
----------
- Sun Jun 25 13:00:00 2017

Modifications
-------------
- Sun Jun 25 13:00:00 2017

Aims
----
- utils for computing in parallel

"""

from copy import deepcopy
import numpy as np
from ipyparallel import Client


def launch_ipcluster_dv(profile="default", targets="all", block=True, max_engines=None):
    # initiate ipcluster
    rc = Client(profile=profile)

    # print ipcluster information
    n_proc = len(rc.ids)
    if targets == "all":
        targets = rc.ids

    dv = rc.direct_view(targets=targets)

    # check number of engines
    # print(rc.ids, dv.targets, targets, max_engines)
    if max_engines is not None:
        if len(dv.targets) > max_engines:
            targets = deepcopy(dv.targets)
            np.random.shuffle(targets)
            targets = targets[:max_engines]
            targets.sort()

            dv = rc.direct_view(targets=targets)

    print("===================================================")
    print("@Slam: ipcluster[{}, n_engines={}/{}]".format(
        profile, len(dv.targets), n_proc))
    print("---------------------------------------------------")

    dv.block = block

    # import basic modules in ipcluster
    dv.execute("import os")
    dv.execute("import numpy as np")
    dv.execute("from joblib import Parallel, delayed, dump, load")

    # print host information
    dv.execute("host_names = os.uname()[1]").get()
    u_host_names, u_counts = np.unique(
        dv["host_names"], return_counts=True)
    for i in range(len(u_counts)):
        print("host: {} x {}".format(u_host_names[i], u_counts[i]))
    print("===================================================")

    return dv


def reset_dv(dv):
    dv.execute("import IPython\n"
               "ipy=IPython.get_ipython()\n"
               "ipy.run_line_magic(\"reset\", \" -f\")\n")
    return


def print_time_cost(dtime, unit_max="hour"):
    """ return string for delta_time """
    if dtime <= 60 * 1.5:
        dtime_str = "{:.3f} sec".format(dtime)
    elif dtime <= 60 * 60 * 1.5:
        dtime_str = "{:.3f} min".format(dtime / 60.)
    elif dtime <= (60 * 60 * 24 * 3):
        dtime_str = "{:.3f} hours".format(dtime / 3600.)
    else:
        # even larger
        if unit_max == "hour":
            dtime <= (60 * 60 * 24 * 3)
            dtime_str = "{:.3f} hours".format(dtime / 3600.)
        else:
            dtime_str = "{:.3f} days".format(dtime / 86400.)

    return dtime_str





# -*- coding: utf-8 -*-
"""

Author
------
Bo Zhang

Email
-----
bozhang@nao.cas.cn

Created on
----------
- Thu Feb 16 17:00:00 2016

Modifications
-------------
- Thu Feb 16 17:00:00 2016

Aims
----
- SlamModel

"""


from abc import ABCMeta, abstractmethod

import numpy as np
from sklearn.neural_network import MLPRegressor
from sklearn.svm import SVR, NuSVR
from sklearn.tree import DecisionTreeRegressor
from sklearn.model_selection import GridSearchCV
from sklearn import ensemble
from sklearn.model_selection import cross_val_score
from sklearn.metrics import mean_squared_error


METHOD_ALL = ["simple", "grid"]

MODEL_ALL = ["svr", "nusvr", "nn", "dt"]
MODEL_MAP = {'svr': SVR,
             'nusvr': NuSVR,
             'nn': MLPRegressor,
             # the three types above are recommended
             'dt': DecisionTreeRegressor,
             'abr': ensemble.AdaBoostRegressor,
             'br': ensemble.BaggingRegressor,
             'etr': ensemble.ExtraTreesRegressor,
             'gbr': ensemble.GradientBoostingRegressor,
             'rfr': ensemble.RandomForestRegressor
             }

SCORING_ALL = ["neg_mean_absolute_error",
               "neg_mean_squared_error",
               "neg_median_absolute_error"
               "r2"]


class SlamModel(object):

    regressor = None
    model = "svr"
    method = ""
    cv = 1
    scoring = "neg_mean_squared_error"

    # trained
    trained = False

    # GridSearchCV attributes
    cv_results_ = None

    def __init__(self, model="nn", method="grid",
                 param_grid=None, cv=8, scoring="neg_mean_squared_error",
                 **kwargs):

        try:
            assert model in MODEL_ALL
        except AssertionError as ae:
            print("@SlamModel: invalid kind!")
            raise ae

        try:
            assert method in METHOD_ALL
        except AssertionError as ae:
            print("@SlamModel: invalid method!")
            raise ae

        self.model = model
        self.method = method
        self.param_grid = param_grid
        self.cv = np.int(cv)
        self.scoring = scoring
        self.score_ = 0

        if self.method == "simple":
            self.regressor = MODEL_MAP[model](**kwargs)
        elif self.method == "grid":
            assert param_grid is not None
            assert self.cv > 2
            assert self.scoring in SCORING_ALL

            self.regressor = GridSearchCV(MODEL_MAP[model](**kwargs),
                                          cv=self.cv, param_grid=param_grid,
                                          scoring=self.scoring)

    def update(self):
        sm = SlamModel()
        sm.regressor = self.regressor
        sm.model = self.model
        sm.method = self.method
        sm.cv = self.cv
        sm.scoring = self.scoring
        sm.trained = self.trained
        sm.cv_results_ = self.cv_results_

        return sm

    def eval_score(self, X, y, sample_weight=None):
        if self.method == "grid":
            self.score_ = self.best_score_
        else:
            if self.model == "nn":
                self.score_ = self.regressor.score(X, y)
            else:
                self.score_ = self.regressor.score(X, y, sample_weight)

    def score(self, *args, **kwargs):
        return self.regressor.score(*args, **kwargs)

    def cross_val_score(self, X, y):
        """ retrun NMSE if cv<2 """
        if self.cv < 2:
            return mean_squared_error(y, self.predict(X), None)
        else:
            return cross_val_score(self.regressor, X, y, cv=self.cv,
                                   scoring=self.scoring).mean()

    def fit(self, X, y, weight=None):
        mock = False
        if weight is None:
            # support weight
            if self.model == "nn":
                self.regressor.fit(X, y)
            else:
                self.regressor.fit(X, y)

            if self.method == "grid":
                self.score_ = self.regressor.best_score_
            else:
                self.score_ = self.score(X, y)

        else:
            ind_weight = weight > 0
            X_ = X[ind_weight]
            y_ = y[ind_weight]
            weight_ = weight[ind_weight]

            if np.sum(ind_weight) < self.cv:
                X_ = np.zeros((self.cv, X.shape[1]), float)
                y_ = np.zeros((self.cv,), float)
                weight_ = np.zeros((self.cv,), float)
                mock = True

            # support weight
            if self.model == "nn":
                self.regressor.fit(X_, y_)
                if self.method == "grid":
                    self.score_ = self.regressor.best_score_
                else:
                    self.score_ = self.cross_val_score(X_, y_)

            else:
                self.regressor.fit(X_, y_, weight_)

                if self.method == "grid":
                    self.score_ = self.regressor.best_score_
                else:
                    self.score_ = self.cross_val_score(X_, y_, weight_)

        self.trained = True
        if mock:
            self.score_ = 1.0

        return

    def predict(self, X):
        return self.regressor.predict(X)

    @staticmethod
    def train(X, y, sample_weight=None, model="nn", method="grid",
              param_grid=None, cv=8, scoring="neg_mean_squared_error",
              **kwargs):
        """ train a single pixel using GridSearchCV

        Parameters
        ----------
        X: ndarray with shape (n_obs x n_dim)
            X in sklearn notation
        y: ndarray with shape (n_obs, ) --> 1D
            y in sklearn notation
        sample_weight: ndarray with shape (n_obs, ) --> 1D
            weight for sample data
        model:
            model type
        method:
            "simple" | "grid"
        param_grid: dict
            key, value pairs of hyper-parameter grids
            >>> param_grid = dict(C=2. ** np.arange(-5., 6.),
            >>>                   epsilon=[0.01, 0.05, 0.1, 0.15],
            >>>                   gamma=['auto', 0.2, 0.25, 0.3, 0.5])
        cv: int / None
            if cv>=3, Cross-Validation will be performed to calculate MSE
        scoring:
            the scoring scheme of cross validation
        kwargs:
            extra kwargs will be passed to svm.SVR() method
            e.g., C=1.0, gamma='auto', epsilon=0.1

        Returns
        -------
        svm.SVR() instance & best hyper-parameters & score
        if CV is not performed, score = np.nan

        """
        sm = SlamModel(model=model, method=method,
                       param_grid=param_grid, cv=cv, scoring=scoring,
                       **kwargs)

        # fit data
        sm.fit(X, y, sample_weight)

        return sm, sm.score_


class Model(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def train(self):
        pass

    @abstractmethod
    def predict_single_spectrum(self):
        pass

    @abstractmethod
    def predict_multi_spectra(self):
        pass


def nmse(model, X, y, sample_weight=None):
    """ return NMSE for svr, X, y and sample_weight """
    if sample_weight is None:
        sample_weight = np.ones_like(y, int)

    ind_use = sample_weight > 0
    if np.sum(ind_use) > 0:
        X_ = X[ind_use]
        y_ = y[ind_use]
        # sample_weight_ = sample_weight[ind_use]
        return -np.mean(np.square(model.predict(X_) - y_))
    else:
        return 0.


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='astroslam',
    version="1.2019.0109.0",
    author='Bo Zhang',
    author_email='bozhang@nao.cas.cn',
    description=('A forward model using SVR to estimate stellar parameters'
                 ' from spectra.'),  # short description
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='http://github.com/hypergravity/astroslam',
    packages=setuptools.find_packages(),
    license='MIT',
    classifiers=["Development Status :: 5 - Production/Stable",
                 "Intended Audience :: Science/Research",
                 "License :: OSI Approved :: MIT License",
                 "Operating System :: OS Independent",
                 "Programming Language :: Python :: 3.7",
                 "Topic :: Scientific/Engineering :: Physics",
                 "Topic :: Scientific/Engineering :: Astronomy"],
    package_dir={'slam': 'slam',
                 'extern': 'slam/extern'},
    package_data={'slam': ['data/*.csv']},
    # include_package_data=True,
    requires=['numpy', 'scipy', 'matplotlib', 'astropy', 'sklearn', 'joblib',
              'pandas', 'emcee', 'lmfit', 'ipyparallel']
)
